#------------------------------------------------------------------
# Copyright(C) 1998-2006  VERITAS Software Corporation.
#
# ALL RIGHTS RESERVED. UNPUBLISHED --  RIGHTS  RESERVED  UNDER  THE
# COPYRIGHT LAWS OF THE UNITED STATES. USE OF A COPYRIGHT NOTICE IS
# PRECAUTIONARY ONLY AND DOES NOT IMPLY PUBLICATION OR DISCLOSURE.
#
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS
# OF   VERITAS   SOFTWARE.  USE,  DISCLOSURE,  OR  REPRODUCTION  IS
# PROHIBITED  WITHOUT  THE  PRIOR  EXPRESS  WRITTEN  PERMISSION  OF
# VERITAS SOFTWARE.
#
#                     RESTRICTED RIGHTS LEGEND
#
# USE, DUPLICATION, OR DISCLOSURE BY THE GOVERNMENT IS  SUBJECT  TO
# RESTRICTIONS  AS  SET  FORTH  IN SUBPARAGRAPH (C) (1) (ii) OF THE
# RIGHTS IN TECHNICAL DATA AND COMPUTER SOFTWARE  CLAUSE  AT  DFARS
# 252.227-7013.
#
# VERITAS  SOFTWARE  350  ELLIS  STREET,  MOUNTAIN  VIEW, CA  94043
#
# $Header:
#
# $Author:
# $Date:
# $Revision:
#
#                                      Veritas Software Corporation
#
# Provides:                            SunJESAdm
# Author:                              Anshuman Atre
# Creation Date:                       21 Nov, 2005
# Modification History:
# Known bugs/caveats:
#------------------------------------------------------------------

#------------
# Be strict..
#------------
use strict;

#----------------------
# Declare the package..
#----------------------
package SunJESAdm;

#---------------------------------
# Include standard PERL packages..
#---------------------------------
use Fcntl;
use Socket;
use POSIX qw ( tmpnam );

#----------------------------------------------------
# Hard code if you still didn't get it from the env..
#----------------------------------------------------
my ( $CLUS_HOME, $sPrefix ) = ();

if ( $ENV { CLUSTER_PRODUCT } &&
     $ENV { CLUSTER_PRODUCT } eq "VAD" ) { # VAD..
  $CLUS_HOME ||= $ENV { VAD_HOME } ||= "/opt/VRTSvad";
  $sPrefix = 'vad';
}
else { # VCS
  $CLUS_HOME ||= $ENV { VCS_HOME } ||= "/opt/VRTSvcs";
  $sPrefix = 'vcs';
}

#-----------------------------------------
# Define module-vendor specific commands..
#-----------------------------------------
my $sCmdHares  = "$CLUS_HOME/bin/hares";
my $sCmdHatype = "$CLUS_HOME/bin/hatype";
my $sCmdVcsDecrypt = "$CLUS_HOME/sbin/${sPrefix}decrypt";

#-------------------------------
# Define list to use in online..
#-------------------------------
my @lsRunningProcs = (); # Global list..

#-------------------------------------
# Include required framework modules..
#-------------------------------------
use Sys;            # support generalised UNIX commands
use Proc;           # For calls to the process routines
use Comms;          # For calls to socket(3C) routines
use Version;        # Version display generator
use VCSagentFW;     # Primary Agent Interface to the ACC Library
use VCSDefines;     # This contains constants that VCS uses

#-----------------------------------------------------------------------
# Add our version number to the version hash and check if versions are
# requested..
#-----------------------------------------------------------------------
my $sVer = "\$Id: SunJESAdm.pm,v 1.26 2006/04/20 09:47:44 aatre Exp $::ETVHeader";
Version::TellVersion ( $sVer );

#--------------------------------
# Define SunJESAdm subroutines..
#--------------------------------

=pod

 -------------------------------------------------------------------------
 Function :  bool ArgsValid ( IN String EntryPointName,
                              IN String ResourceName,
                              IN String AdminHost,
                              IN Integer AdminPort,
                              IN String AdminUser,
                              IN String AdminPasswd,
                              IN String MonitorProgram,
                              IN Integer SecondLevelMonitor,
                              IN String ServerRoot,
                              IN String SSLDbPasswd,
                              IN Boolean SSLEnabled );

 Description : Validates the ArgList passed to the entry point.

 Parameters :

 $sEntryPointName           Entry pt name
 $sResourceName             Name of this VCS resource
 $localAdminHost            Hostname/IPv4 address of virtual host for Admin Server
 $localAdminPort            The port at which the Admin Server listens
 $localAdminUser            The administrative user for the Admin Server
 $localAdminPasswd          Encrypted password for AdminUser using vcsencrypt
 $localMonitorProgram       Full path to external monitor program
 $localSecondLevelMonitor   Cycle count to run in-depth monitoring
 $localServerRoot           Absolute path to ServerRoot of Admin Server
 $localSSLDbPasswd          Encrypted password for trusted database using vcsencrypt
 $localSSLEnabled           Flag informing that the server uses SSL
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub ArgsValid {
  #----------------------------------------
  # Set up local vars from passed in ones..
  #----------------------------------------
  my ( $sEntryPointName,
       $sResourceName,
       $localAdminHost,
       $localAdminPort,
       $localAdminUser,
       $localAdminPasswd,
       $localMonitorProgram,
       $localSecondLevelMonitor,
       $localServerRoot,
       $localSSLDbPasswd,
       $localSSLEnabled ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20001, "Subroutine <%s> called with:
  EntryPointName           [%s]
  ResourceName             [%s]
  AdminHost                [%s]
  AdminPort                [%s]
  AdminUser                [%s]
  AdminPasswd              [%s]
  MonitorProgram           [%s]
  SecondLevelMonitor       [%s]
  ServerRoot               [%s]
  SSLDbPasswd              [%s]
  SSLEnabled               [%s]",

  $sMyName,
  $sEntryPointName, $sResourceName, $localAdminHost, $localAdminPort,
  $localAdminUser,       $localAdminPasswd,     $localMonitorProgram,
  $localSecondLevelMonitor,    $localServerRoot,   $localSSLDbPasswd,
  $localSSLEnabled );

  #-------------------------------------
  # Set return code to hopeful success..
  #-------------------------------------
  my $iReturnCode = 1;

  #--------------------------------------------
  # Create hash to weed out empty args values..
  #--------------------------------------------
  my ( @lsMissing, %hsArgList ) = ();
  my @lsArgs =
  qw ( EntryPointName ResourceName AdminHost   AdminPort AdminUser
       AdminPasswd MonitorProgram   SecondLevelMonitor  ServerRoot
       SSLDbPasswd SSLEnabled );

  @hsArgList { @lsArgs } = @_; # Assign..

  #----------------------------------------
  # Reqd. attributes based on entry point..
  #----------------------------------------
  my @lsOnlineArgs  = qw ( AdminHost AdminPort ServerRoot );
  my @lsOfflineArgs = qw ( ServerRoot );
  my @lsMonitorArgs = qw ( ServerRoot );
  my @lsCleanArgs   = qw ( ServerRoot );

  my %hlsNeeded = ( 'online'  => \@lsOnlineArgs,
                    'offline' => \@lsOfflineArgs,
                    'monitor' => \@lsMonitorArgs,
                    'clean'   => \@lsCleanArgs );

  my @lsReqdArgs = @{$hlsNeeded { $sEntryPointName }};

  #-----------------------
  # Check what's missing..
  #-----------------------
  if ( @lsMissing = grep { $hsArgList { $_ } !~ /\S/o } @lsReqdArgs ) {
    &logError ( 20002,
    "<%s> The following mandatory fields were not set [%s], Total [%s]",
    $sMyName, join ( ' ', @lsMissing ), scalar @lsMissing );
    $iReturnCode = 0; # Failure..
  }

  #----------------------------
  # Get current resource type..
  #----------------------------
  my $sResourceType = ( $0 =~ m@([^/]+)@g ) [-2];

  #------------------------------------------
  # Force SLM if enabled, for next monitor..
  #------------------------------------------
  if ( $sEntryPointName ne 'monitor' ) {
    SunJESAdm::RunSecondLevelMonitor ( 1, $sResourceType, $sResourceName );
  }

  #----------------------------------------
  # Return, if failed or monitor or clean..
  #----------------------------------------
  if ( ! $iReturnCode                  ||
         $sEntryPointName eq 'monitor' ||
         $sEntryPointName eq 'clean' ) {
    return $iReturnCode;
  }

  #-------------------------------
  # Individual validation starts..
  #-------------------------------

  #-------------------------
  # Server root must exist..
  #-------------------------
  if ( ! -d $localServerRoot ) {
    &logError ( 20003, "<%s> ServerRoot directory [%s] does not exist",
    $sMyName, $localServerRoot );
    $iReturnCode = 0; # Failure..
  }

  if ( $sEntryPointName eq 'online' ) {
    #--------------------
    # Validate hostname..
    #--------------------
    if ( ! inet_aton ( $localAdminHost ) ) {
      &logError ( 20004, "<%s> Could not resolve AdminHost [%s]",
      $sMyName, $localAdminHost );
      $iReturnCode = 0; # Failure.. 
    }
  
    #-------------
    # Admin Port..
    #-------------
    if ( $localAdminPort <= 0 ) {
      &logError ( 20005, "<%s> AdminPort [%s] has to be greater than 0",
      $sMyName, $localAdminPort );
      $iReturnCode = 0; # Failure.. 
    }

    #--------------------------------------
    # MonitorProgram has to be executable..
    #--------------------------------------
    if ( $localMonitorProgram =~ /(\S+)/ &&     # Defined..
         ! -x $1 ) {                            # And not executable..
      &logError ( 20006,
      "<%s> MonitorProgram [%s] not executable", $sMyName, $1 );
      $iReturnCode = 0; # Failure..
    }

    #--------------
    # AdminPasswd..
    #--------------
    if ( $localAdminUser =~ /\S/ &&    # Defined..
         $localAdminPasswd !~ /\S/ ) { # And not defined..
      &logError ( 20007,
      "<%s> AdminPasswd cannot be NULL [%s] if AdminUser [%s] is defined",
      $sMyName, $localAdminPasswd, $localAdminUser );
      $iReturnCode = 0; # Failure..
    }
  }

  $iReturnCode ||
  &logWarn ( 20008,
  "<%s> Some of the attributes could not be validated. Returning [%s]",
  $sMyName, $iReturnCode );

  return $iReturnCode ;
}

=pod

 -------------------------------------------------------------------------
 Function : bool Start ( IN String SSLDbPasswd,
                         IN String ServerRoot );

 Description : Start logic needed for online entry pt

 Parameters :

 $sSSLDbPasswd      Encrypted password for trusted database using vcsencrypt
 $sServerRoot       Absolute path to ServerRoot of Admin Server
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Start {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sSSLDbPasswd,
       $sServerRoot ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20009, "Subroutine <%s> called with:
  SSLDbPasswd              [%s]
  ServerRoot               [%s]",

  $sMyName, $sSSLDbPasswd, $sServerRoot );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #-----------------------------
  # Initialise local variables..
  #-----------------------------
  my ( $iTillWhen, $iTimeout, @liDummyPids ) = ();
  my $iSleepTime = 10; # Secs..

  #--------------------------------
  # Kill any process if remaining..
  #--------------------------------
  if ( SunJESAdm::KillAdminProcs ( $sServerRoot ) ) {
    &logTrace ( 20010,
    "<%s> Confirmed no persistent processes exist before startup",
    $sMyName );
  }
  else {
    &logError ( 20011,
    "<%s> Persistent processes exist for Administration server", $sMyName );
  }

  #----------------------------------
  # Check if SSLDbPasswd is defined..
  #----------------------------------
  if ( $sSSLDbPasswd =~ /\S/ ) { # Not null..
    &logTrace ( 20012,
    "<%s> Starting server using crypted protected database passwd [%s]",
    $sMyName, $sSSLDbPasswd );
  }
  else {
    &logTrace ( 20013, "<%s> Starting server without protected database passwd", 
    $sMyName );
  }

  #-------------------------
  # Start the admin server..
  #-------------------------
  if ( SunJESAdm::ExecAdmCmd ( $sServerRoot, $sSSLDbPasswd ) ) {
    &logTrace ( 20014, "<%s> Started the Administration Server successfully!",
    $sMyName );

    #--------------------
    # Delay entry point..
    #--------------------
    $iTimeout = SunJESAdm::GetEngineTimeout ( 1 ); # With buffer..

    while ( ( $iTillWhen = $^T + $iTimeout - time )  > 10 ) {

      if ( VCSDef_MonitorExit::ONLINE ==
           SunJESAdm::GetProcState ( $sServerRoot, \@liDummyPids, 1 ) ) {
        &logInfo ( 20015,
        "<%s> Confirmed startup of Administration Server with PIDs [%s]",
        $sMyName, join ( ' ', @liDummyPids ) );
        last;
      }
 
      &logTrace ( 20016, "<%s> Could not confirm startup.
      Delaying exit with [%s] sec in hand", $sMyName, $iTillWhen );

      sleep $iSleepTime; # Pause before re-trying..
    }
  }
  else {
    &logError ( 20017, "<%s> Could NOT start the Administration Server",
    $sMyName );
    $iReturnCode = 0; # Failure..
  }

  return $iReturnCode;
}

=pod

 -------------------------------------------------------------------------
 Function : bool Stop ( IN String ServerRoot );

 Description : Stop logic needed for offline entry pt

 Parameters :

 $sServerRoot               Absolute path to ServerRoot of Admin Server
 

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Stop {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sServerRoot ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20018, "Subroutine <%s> called with:
  ServerRoot               [%s]",
  $sMyName, $sServerRoot );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1; # Success..

  #-----------------------------
  # Initialise local variables..
  #-----------------------------
  my ( $iTillWhen, $iTimeout, @liDummyPids ) = ();
  my $iSleepTime = 10; # Secs..

  #----------------------------
  # Check if PID files exist..
  #----------------------------
  if ( 2 == scalar ( SunJESAdm::GetPidFiles ( $sServerRoot ) ) ) {
    &logTrace ( 20019,
    "<%s> All PID files located for Administration Server", $sMyName );
  }
  else {
    &logWarn ( 20020, "<%s> Some PID files have problems", $sMyName );
  }

  #------------------------
  # Stop the admin server..
  #------------------------
  if ( SunJESAdm::ExecAdmCmd ( $sServerRoot, "" ) ) {
    &logTrace ( 20021, "<%s> Stopped the Administration Server successfully!",
    $sMyName );

    #--------------------
    # Delay entry point..
    #--------------------
    $iTimeout = SunJESAdm::GetEngineTimeout ( 1 ); # With buffer..

    while ( ( $iTillWhen = $^T + $iTimeout - time ) > 10 ) {

      if ( VCSDef_MonitorExit::OFFLINE ==
           SunJESAdm::GetProcState ( $sServerRoot, \@liDummyPids, 0 ) ) {
        &logInfo ( 20022,
        "<%s> Confirmed shutdown of Administration Server", $sMyName );
        last;
      }

      &logTrace ( 20023,
      "<%s> Could not confirm shutdown. Retrying in [%s] sec.
       Delaying kill with [%s] sec in hand",
       $sMyName, $iSleepTime, $iTillWhen );

      sleep $iSleepTime; # Pause before re-trying..
    }

  }
  else {
    &logError ( 20024, "<%s> Could NOT stop the Administration Server",
    $sMyName );
    $iReturnCode = 0; # Failure..
  }

  #--------------------------------
  # Kill any process if remaining..
  #--------------------------------
  if ( SunJESAdm::KillAdminProcs ( $sServerRoot ) ) {
    &logTrace ( 20025,
    "<%s> Confirmed no pending processes exist after shutdown",
    $sMyName );
  }
  else {
    &logError ( 20026,
    "<%s> Pending processes exist for Administration server", $sMyName );
    $iReturnCode = 0; # Failure..
  }

  return $iReturnCode;
}

=pod

 -------------------------------------------------------------------------
 Function : bool Kill ( IN String ServerRoot );

 Description : Forces shutdown of resource needed for clean entry point

 Parameters :

 $sServerRoot               Absolute path to ServerRoot of Admin Server
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Kill {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sServerRoot ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20027, "Subroutine <%s> called with:
  ServerRoot               [%s]",
  $sMyName, $sServerRoot );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #-----------------------------------
  # Attempt to stop the admin server..
  #-----------------------------------
  if ( SunJESAdm::ExecAdmCmd ( $sServerRoot, "" ) ) {
    &logTrace ( 20028, "<%s> Stopped the Administration Server successfully!",
    $sMyName );
  }
  else {
    &logError ( 20029, "<%s> Could NOT stop the Administration Server",
    $sMyName );
  }

  #-----------------------------
  # Kill any pending processes..
  #-----------------------------
  if ( SunJESAdm::KillAdminProcs ( $sServerRoot ) ) {
    &logTrace ( 20030, "<%s> Confirmed no pending processes exist",
    $sMyName );
  }
  else {
    &logWarn ( 20031,
    "<%s> Pending processes exist for Administration server", $sMyName );
  }

  return $iReturnCode;
}

=pod

 -------------------------------------------------------------------------
 Function : int GetResourceState ( IN String AdminHost,
                                   IN String ServerRoot );

 Description : First level monitor to determine state of resource.

 Parameters :

 $sAdminHost       Hostname/IPv4 address of virtual host for Admin Server
 $sServerRoot      Absolute path to ServerRoot of Admin Server
 

 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 -------------------------------------------------------------------------

=cut

sub GetResourceState {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sAdminHost,
       $sServerRoot ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20032, "Subroutine <%s> called:
  AdminHost                [%s]
  ServerRoot               [%s]",

  $sMyName, $sAdminHost, $sServerRoot );

  my @liDummyPids = ();

  #---------------------------------
  # Check if IP is plumbed locally..
  #---------------------------------
  unless ( SunJESAdm::IsIPLocal ( $sAdminHost ) ) {
    &logTrace ( 20033,
    "<%s> Virtual Host [%s] not up locally",
    $sMyName, $sAdminHost );
    return VCSDef_MonitorExit::OFFLINE;
  }

  my $iReturnCode = SunJESAdm::GetProcState ( $sServerRoot,
                                              \@liDummyPids,
                                              1 ); # Try fast FLM..

  if( VCSDef_MonitorExit::ONLINE == $iReturnCode ) {
    &logTrace ( 20034, "<%s> Found all required processes",
    $sMyName );
  }
  else {
    &logTrace ( 20035,
    "<%s> did not find all required processes, returning [%s]",
    $sMyName, $iReturnCode );
  }

  @lsRunningProcs = @liDummyPids; # Populate global..

  &logTrace ( 20036,
  "<%s> Filling global array with [%s] elements and Returning [%s]",
  $sMyName, scalar @lsRunningProcs, $iReturnCode );

  return ( $iReturnCode );
}

=pod

 -------------------------------------------------------------------------
 Function : int GetProcState ( IN String ServerRoot,
                               OUT ListReference OnlinePids,
                               In Boolean UseFFLM );

 Description : Performs the actual first level monitoring

 Parameters :

 $sServerRoot      Absolute path to ServerRoot of Admin Server
 $rliOnlinePids    PIDs of relevant online processes
 $bUseFFLM         Flag to indicate if fast FLM is to be attempted 

 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 -------------------------------------------------------------------------

=cut

sub GetProcState {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sServerRoot,
       $rliOnlinePids,
       $bUseFFLM ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20037, "Subroutine <%s> called with:
  ServerRoot        [%s]
  rliOnlinePids     [%s]
  UseFFLM           [%s]",

  $sMyName, $sServerRoot, $rliOnlinePids, $bUseFFLM );

  #----------------------------------------
  # Set Return Code Optimistically to 110..
  #----------------------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE;

  #----------------------
  # Initialise Variables..
  #----------------------
  my ( $iProcessCount, $rhlProc, $iPid, $iExitCode, $iWDogPid,
       $sPsCmd, $sLine, $sPidFile, $sPattern, $sPattern1, @lsPidFiles,
       @iNsPPids, @hProcListHash, %hsPidfilePtrn ) = ();

  my $sFmt = ".*$sServerRoot/admin-serv/config";
  my @lsProcPtrn = qw ( uxwdog ns-httpd );         # In this order..
  my $iFound = 0; # Flag to indicate ns-httpd was located..
  @$rliOnlinePids = (); # Reset..
 
  #---------------------
  # Attempt to be fast..
  #---------------------
  if ( $bUseFFLM ) {
    if ( 2 == scalar ( @lsPidFiles =
              SunJESAdm::GetPidFiles ( $sServerRoot ) ) ) {
  
      &logTrace ( 20038, "<%s> Located both PID files. Attempting fast FLM.",
      $sMyName );
  
      $sPsCmd = "/bin/ps";
      @hsPidfilePtrn { @lsPidFiles } = @lsProcPtrn; # Assign..
  
      #--------------------------------------
      # Search the PIDs using the PID files..
      #--------------------------------------
      foreach $sPidFile ( keys %hsPidfilePtrn ) {
        if ( open ( PIDFILE, $sPidFile ) ) {
          ( $iPid = <PIDFILE> ) =~ s/\n//; # Read a line.. and chomp..
  
          if ( $iPid && $iPid =~ /\d+/ ) { # Valid PID..
            &logTrace ( 20039, "<%s> +--- Got PID [%s] from PID file [%s] ---+",
            $sMyName, $iPid, $sPidFile );
  
            close ( PIDFILE ) ||
            &logWarn ( 20040, "<%s> Could not close PID file [%s] after read: [%s]",
            $sMyName, $sPidFile, $! );
  
            if ( kill ( 0, $iPid ) ) { # Not a defunct..
              ( $sLine ) = grep { s/\n// || 1 } # Chomp..
                           readpipe ( "$sPsCmd -p $iPid -o args=" );
  
              #--------------------------
              # Check command execution..
              #--------------------------
              if ( $iExitCode = $? >> 8 ) { # Get errno..
                &logError ( 20041,
                "<%s> ps(1m) command [%s] for PID [%s] failed with exit code [%s]",
                $sMyName, $sPsCmd, $iPid, $iExitCode );
                @lsPidFiles = (); # Reset..
                last; # Get out.. 
              }
              else { # Success..
                #-----------------------
                # Create search format..
                #-----------------------
                $sPattern = $hsPidfilePtrn { $sPidFile } . $sFmt; # Join..
  
                &logTrace ( 20042,
                "<%s> Using search filter [%s] on CmdArgs [%s] PID [%s]",
                $sMyName, $sPattern, $sLine, $iPid );
              }
  
              #-------------------------------
              # Check if we got what we want..
              #-------------------------------
              if ( $sLine =~ /$sPattern\b/ ) {                       # Found..
                if ( $hsPidfilePtrn { $sPidFile } eq 'ns-httpd' ) {  # ns-httpd..
                  &logTrace ( 20043,
                  "<%s> +--- Found [%s] process with PID [%s] CmdArgs [%s] ---+",
                  $sMyName, $hsPidfilePtrn { $sPidFile }, $iPid, $sLine );
  
                  push ( @$rliOnlinePids, $iPid ); # Store..
                  ++ $iFound;
                }
                elsif ( $hsPidfilePtrn { $sPidFile } eq 'uxwdog' ) { # Watchdog..
                  &logTrace ( 20044,
                  "<%s> Found watchdog [%s] process with PID [%s] CmdArgs [%s]",
                  $sMyName, $hsPidfilePtrn { $sPidFile }, $iPid, $sLine );
  
                  $iWDogPid = $iPid;
                  unshift ( @$rliOnlinePids, $iPid ); # Store..
                }
              }
              else { # Oops! Stale PID!?
                &logWarn ( 20045,
                "<%s> Encountered stale PID [%s] with Args [%s] from PID file [%s]",
                $sMyName, $iPid, $sLine, $sPidFile );
  
                @lsPidFiles = (); # Reset..
                last; # Get out.. 
              }
  
              #----------------------------------------
              # Return if we got the ns-httpd process..
              #----------------------------------------
              if ( $iFound &&
                   $iWDogPid ) {
                &logTrace ( 20046,
                "<%s> Located ns-httpd process. Returning [%s] from fast FLM",
                $sMyName, $iReturnCode );
  
                return $iReturnCode; # Get out..
              }
            }
            else {
              &logTrace ( 20047,
              "<%s> PID [%s] from PID file [%s] does not exist",
              $sMyName, $iPid, $sPidFile );
  
              @lsPidFiles = (); # Reset..
              last; # Get out.. 
            }
          }
          else { # Corrupted..
            &logWarn ( 20048, "<%s> PID file [%s] seems corrupted",
            $sMyName, $sPidFile );
            @lsPidFiles = (); # Reset..
            last;            # Get out..
          }
        }
        else {
          &logWarn ( 20049, "<%s> Problems opening PID file [%s] for read: [%s]",
          $sMyName, $sPidFile, $! );
          @lsPidFiles = (); # Reset..
          last;            # Get out..
        }
      }
    }
    else {
      &logTrace ( 20050, "<%s> Got only [%s] PID files",
      $sMyName, scalar @lsPidFiles );
      @lsPidFiles = (); # Reset..
    }
  
    return $iReturnCode if scalar @lsPidFiles; # Is this needed..?
  
    &logTrace ( 20051,
    "<%s> Could not locate both PID files. Attempting regular FLM..",
    $sMyName );
  }

  @$rliOnlinePids = (); # Reset..
  #-----------------------------------
  # Call processes and store in hash..
  #-----------------------------------
  $iProcessCount = Proc::GetProcessListHash ( \@hProcListHash );

  #----------------------
  # Check if we got any..
  #----------------------
  if ( 0 == $iProcessCount ) { # Oops!
    &logError ( 20052,
    "<%s> Call to get process list did not return any processes",
    $sMyName );
    return VCSDef_MonitorExit::OFFLINE; # Offline..
  }

  #-----------------
  # Create filters..
  #-----------------
  $sPattern  = "ns-httpd${sFmt}";
  $sPattern1 = "uxwdog${sFmt}";

  &logTrace ( 20053, "<%s>
  -----------------------------------------------
  Got [%s] processes from proc table. Using regex
  [%s] and\n  [%s]
  for filtering
  -----------------------------------------------",
  $sMyName, scalar @hProcListHash, $sPattern, $sPattern1 );

  #-----------------------
  # Search for processes..
  #-----------------------
  foreach $rhlProc ( @hProcListHash ) {
    #--------------------------------------------------------------
    # Each $rhlProc is a ref. to a hash with the following keys..
    # "UID"     =>  <user id>
    # "PID"     =>  <process id>
    # "PPID"    =>  <parent process id>
    # "COMMAND" =>  <command string>
    #--------------------------------------------------------------
  
    &logTrace ( 20054, "<%s> Searching PID [%s] CMD [%s]",
    $sMyName, $$rhlProc{"PID"}, $$rhlProc{"COMMAND"} );

    if ( $$rhlProc{"COMMAND"} =~ /$sPattern\b/o ) { # Found ns-httpd..
      &logTrace ( 20055,
      "<%s> +--- FOUND [ns-httpd] process with PID [%s] Cmdargs [%s] ---+",
      $sMyName, $$rhlProc{"PID"}, $$rhlProc{"COMMAND"} );

      push ( @iNsPPids, $$rhlProc{"PPID"} );      # Store..
      push ( @$rliOnlinePids, $$rhlProc{"PID"} ); # Store..
      ++ $iFound;
    }
    elsif ( $$rhlProc{"COMMAND"} =~ /$sPattern1\b/o ) { # Found uxwdog..
      &logTrace ( 20056,
      "<%s> FOUND [uxwdog] process with PID [%s] CmdArgs [%s]",
      $sMyName,  $$rhlProc{"PID"}, $$rhlProc{"COMMAND"});

      $iWDogPid = $$rhlProc{"PID"};
      unshift ( @$rliOnlinePids, $$rhlProc{"PID"} ); # Store..
    }
    else { # Junk..
      last if ( $iWDogPid && 2 == $iFound );       # Found all..
    }
  }

  #------
  # Log..
  #------
  if ( $iFound ) { # At least one ns-httpd found..
    if ( $iWDogPid &&                        # Got watchdog..
      grep { $iWDogPid == $_ } @iNsPPids ) { # Got parent ns-httpd..
      &logTrace ( 20057, "<%s> Located both processes. Returning [%s]",
      $sMyName, $iReturnCode );
    }
    elsif ( grep { 1 == $_ } @iNsPPids ) {   # Parent belongs to init..
      &logWarn ( 20058,
      "<%s> Watchdog process [uxwdog] is missing, but returning [%s]",
      $sMyName, $iReturnCode );
    }
    else { # Missing parent. Error..
      $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
      &logError ( 20059,
      "<%s> Missing uxwdog. Could not locate parent ns-httpd process.
      Returning [%s]", $sMyName, $iReturnCode );
    }
  }
  else {
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
    &logTrace ( 20060,
    "<%s> Could not locate any ns-httpd processes.
    Returning [%s]", $sMyName, $iReturnCode );
  }

  return $iReturnCode;
}

=pod

 -------------------------------------------------------------------------
 Function : int Monitor ( IN String AdminHost,
                          IN Integer AdminPort,
                          IN String AdminUser,
                          IN String AdminPasswd,
                          IN String MonitorProgram,
                          IN String ServerRoot,
                          IN Integer SecondLevelMonitor,
                          IN Boolean SSLEnabled,
                          IN String ResourceName );

 Description : Performs the in-depth monitoring

 Parameters :

 $sAdminHost            Hostname/IPv4 address of virtual host for Admin Server
 $iAdminPort            The port at which the Admin Server listens
 $sAdminUser            The administrative user for the Admin Server
 $sAdminPasswd          Encrypted password for AdminUser using vcsencrypt
 $sMonitorProgram       Full path to external monitor program
 $sServerRoot           Absolute path to ServerRoot of Admin Server
 $iSecondLevelMonitor   Cycle count to run in-depth monitoring
 $bSSLEnabled           Flag informing that the server uses SSL
 $sResourceName         Name of this VCS resource

 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub Monitor {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sAdminHost,
       $iAdminPort,
       $sAdminUser,
       $sAdminPasswd,
       $sMonitorProgram,
       $sServerRoot,
       $iSecondLevelMonitor,
       $bSSLEnabled,
       $sResourceName ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20061, "Subroutine <%s> called:
  AdminHost             [%s]
  AdminPort             [%s]
  AdminUser             [%s]
  AdminPasswd           [%s]
  MonitorProgram        [%s]
  ServerRoot            [%s]
  SecondLevelMonitor    [%s]
  SSLEnabled            [%s]
  ResourceName          [%s]",

  $sMyName,
  $sAdminHost,   $iAdminPort,   $sAdminUser,   $sAdminPasswd,
  $sMonitorProgram,     $sServerRoot,   $iSecondLevelMonitor,
  $bSSLEnabled, $sResourceName );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..

  #-------------------------------------------------------
  # Upper bound..
  # Time in EPOCH when entry pt started + VCS time limit..
  #-------------------------------------------------------
  my $iTillWhen = $^T + SunJESAdm::GetEngineTimeout ( 1 ); # With margin..

  #----------------------------
  # Get current resource type..
  #----------------------------
  my $sResourceType = ( $0 =~ m@([^/]+)@g ) [-2];
  $sResourceType = 'SunJESAdm' if $sResourceType !~ /\w/;

  #----------------------------------------------
  # In-depth monitoring, if correct time to run..
  #---------------------------------------------- 
  if ( SunJESAdm::RunSecondLevelMonitor ( $iSecondLevelMonitor,
                                          $sResourceType,
                                          $sResourceName ) ) {
    #--------------------------
    # Just In Time Validation..
    #--------------------------
    if ( VCSDef_MonitorExit::ONLINE == ( $iReturnCode =
         SunJESAdm::DelayedAttributeCheck ( $sAdminHost,
                                            $iAdminPort,
                                            $sAdminUser,
                                            $sAdminPasswd,
                                            $sServerRoot ) ) ) {
      &logTrace ( 20062,
      "<%s> Delayed attribute check completed successfully",
      $sMyName );
    }
    else {
      &logError ( 20063,
      "<%s> Attribute validation failed. Returning [%s]",
      $sMyName, $iReturnCode );
      return $iReturnCode;
    }

    &logTrace ( 20064,
    "<%s> Proceeding with second level monitoring", $sMyName );

    if ( VCSDef_MonitorExit::ONLINE == ( $iReturnCode =
         SunJESAdm::SecondLevelMonitor ( $sAdminHost,
                                         $iAdminPort,
                                         $sAdminUser,
                                         $sAdminPasswd,
                                         $sServerRoot,
                                         $bSSLEnabled,
                                         $iTillWhen ) ) ) {
    #----------------------------------
    # Online after in-depth monitoring..
    #----------------------------------
      &logTrace ( 20065,
      "<%s> Administration Server online with status [%s] " .
      "after second level monitoring",
      $sMyName, $iReturnCode );
    }
    else {
      &logError ( 20066,
      "<%s> Administration Server NOT online. Status [%s] " .
      "after second level monitoring",
      $sMyName, $iReturnCode );
      return $iReturnCode; # Get out..
    }
  }

  #------------------------------------------------
  # Exit if external monitor has not been defined..
  #------------------------------------------------
  return $iReturnCode unless defined $sMonitorProgram;

  #---------------------------------------
  # MonitorProgram defined..
  # Separate the executable and the args..
  #---------------------------------------
  my ( $sExtCmd, $sExtCmdArgs ) = $sMonitorProgram # Separate
                                =~ /(\S+)\s*(.*)/; # Cmd and args..
  $sExtCmdArgs = '' unless defined $sExtCmdArgs;   # Prevent undefs..

  if ( $sExtCmd ) {                                      # Defined..
    if ( VCSDef_MonitorExit::ONLINE == ( $iReturnCode =  # Check status..
         SunJESAdm::ExternalMonitor ( $sExtCmd,          # After execution..
                                      $sExtCmdArgs,
                                      $iTillWhen ) ) ) {
  
      &logTrace ( 20067,
      "<%s> Administration server found online with status [%s]
      after executing external monitor [%s] with args [%s]",
      $sMyName, $iReturnCode, $sExtCmd, $sExtCmdArgs );
    }
    else {
      &logError ( 20068,
      "<%s> Administration Server NOT online with status [%s]
      after executing external monitor [%s] with args [%s]",
      $sMyName, $iReturnCode, $sExtCmd, $sExtCmdArgs );
    }
  }

  return $iReturnCode;
}

=pod

 -------------------------------------------------------------------------
 Function : Int DelayedAttributeCheck ( IN String AdminHost,
                                        IN Integer AdminPort,
                                        IN String AdminUser,
                                        IN String AdminPasswd,
                                        IN String ServerRoot );

 Description : Just In Time Validation

 Parameters :

 $sAdminHost      Hostname/IPv4 address of virtual host for Admin Server
 $iAdminPort      The port at which the Admin Server listens
 $sAdminUser      The administrative user for the Admin Server
 $sAdminPasswd    Encrypted password for AdminUser using vcsencrypt
 $sServerRoot     Absolute path to ServerRoot of Admin Server

 Returns  : Integer value as defined below:

 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub DelayedAttributeCheck {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sAdminHost,
       $iAdminPort,
       $sAdminUser,
       $sAdminPasswd,
       $sServerRoot ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20069, "Subroutine <%s> called:
  AdminHost             [%s]
  AdminPort             [%s]
  AdminUser             [%s]
  AdminPasswd           [%s]
  ServerRoot            [%s]",

  $sMyName,
  $sAdminHost, $iAdminPort, $sAdminUser, $sAdminPasswd,
  $sServerRoot );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..

  #-----------------------------------------
  # Validate each of the attribute in turn..
  #-----------------------------------------

  #------------------------
  # ServerRoot must exist..
  #------------------------
  if ( ! -d $sServerRoot ) {
    &logError ( 20070,
    "<%s> ServerRoot directory [%s] does not exist",
    $sMyName, $sServerRoot );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }

  #--------------------
  # Validate hostname..
  #--------------------
  if ( ! inet_aton ( $sAdminHost ) ) {
    &logError ( 20071,
    "<%s> Could not resolve AdminHost [%s]", $sMyName, $sAdminHost );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }

  #-------------
  # Admin Port..
  #-------------
  if ( $iAdminPort <= 0 ) {
    &logError ( 20072,
    "<%s> AdminPort [%s] has to be greater than 0",
    $sMyName, $iAdminPort );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }

  #--------------
  # AdminPasswd..
  #--------------
  if ( $sAdminUser =~ /\S/ &&   # Defined..
      $sAdminPasswd !~ /\S/ ) { # And not defined..
    &logError ( 20073,
    "<%s> AdminPasswd cannot be NULL [%s] if AdminUser [%s] is defined",
    $sMyName, $sAdminPasswd, $sAdminUser );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }

  return $iReturnCode;
}

=pod

 -------------------------------------------------------------------------
 Function : Int SecondLevelMonitor ( IN String AdminHost,
                                     IN Integer AdminPort,
                                     IN String AdminUser,
                                     IN String AdminPasswd,
                                     IN String ServerRoot,
                                     IN Boolean SSLEnabled,
                                     IN Integer TillWhen );

 Description : In-depth monitoring, if enabled.

 Parameters :

 $sAdminHost      Hostname/IPv4 address of virtual host for Admin Server
 $iAdminPort      The port at which the Admin Server listens
 $sAdminUser      The administrative user for the Admin Server
 $sAdminPasswd    Encrypted password for AdminUser using vcsencrypt
 $sServerRoot     Absolute path to ServerRoot of Admin Server
 $bSSLEnabled     Flag informing that the server uses SSL
 $iTillWhen       Absolute time in EPOCH to engine timeout

 Returns  : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub SecondLevelMonitor {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sAdminHost,
       $iAdminPort,
       $sAdminUser,
       $sAdminPasswd,
       $sServerRoot,
       $bSSLEnabled,
       $iTillWhen ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20074, "Subroutine <%s> called:
  AdminHost             [%s]
  AdminPort             [%s]
  AdminUser             [%s]
  AdminPasswd           [%s]
  ServerRoot            [%s]
  SSLEnabled            [%s]
  TillWhen              [%s]",

  $sMyName,
  $sAdminHost, $iAdminPort, $sAdminUser, $sAdminPasswd,
  $sServerRoot, $bSSLEnabled, $iTillWhen );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..

  #-------------------------
  # Define local variables..
  #-------------------------
  my $sMonitorCmd = "$sServerRoot/bin/admin/admconfig";
  my $sDecryptCmd = $sCmdVcsDecrypt; # Get global..
  my ( $sLine, $sCmd, $sCmdArgs, $sEnc, $iTimeout,
       @lsCmdOut, $iExitCode ) = ();
  
  #---------------------------------
  # Check if User has been defined..
  #---------------------------------
  if ( $sAdminUser =~ /\S/ ) {
    #--------------------------------------
    # Check if vcsdecrypt(1m) can be used..
    #--------------------------------------
    if ( ! -e $sDecryptCmd ) {
      &logError ( 20075, "<%s> Decrypt utility [%s] does not exist",
      $sMyName, $sDecryptCmd );
      $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
    }
    elsif ( ! -x $sDecryptCmd ) {
      &logError ( 20076, "<%s> Decrypt utility [%s] not executable",
      $sMyName, $sDecryptCmd );
      $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
    }
    else {
      &logTrace ( 20077, "<%s> Decrypt utility [%s] located",
      $sMyName, $sDecryptCmd );
    }

    #------------------------------------
    # Check if admconfig can be used..
    #------------------------------------
    if ( ! -e $sMonitorCmd ) {
      &logError ( 20078, "<%s> Monitor utility [%s] does not exist",
      $sMyName, $sMonitorCmd );
      $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
    }
    elsif ( ! -x $sMonitorCmd ) {
      &logError ( 20079, "<%s> Monitor utility [%s] not executable",
      $sMyName, $sMonitorCmd );
      $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
    }
    else {
      &logTrace ( 20080, "<%s> Monitor utility [%s] located",
      $sMyName, $sMonitorCmd );
    }

    #----------------------------------
    # Exit, if not errors encountered..
    #----------------------------------
    unless ( $iReturnCode == VCSDef_MonitorExit::ONLINE ) {
      return $iReturnCode;
    }

    #----------------------------------
    # Create cmd string for execution..
    #----------------------------------
    $sEnc = $bSSLEnabled?'-enc ':''; # Use encrypted protocol..
    $sCmd = "$sDecryptCmd";
    $sCmdArgs = "$sAdminPasswd | $sMonitorCmd $sEnc" .
                "-server $sAdminHost:$iAdminPort -user $sAdminUser -getPort";

    #--------------------
    # Get time bearings..
    #--------------------
    if ( ( $iTimeout = $iTillWhen - ( time + 5 ) ) <= 10 ) {
      &logError ( 20081,
      "<%s> Insufficient time to do in-depth monitoring. Short by [%s] sec",
      $sMyName, abs ( $iTimeout ) );
      return VCSDef_MonitorExit::OFFLINE; # Error..
    }

    &logTrace ( 20082,
    "<%s> Executing command [%s] with args [%s] within [%s] secs",
    $sMyName, $sCmd, $sCmdArgs, $iTimeout );
  
    #----------------------
    # Execute the command..
    #----------------------
    if ( SunJESAdm::ExecLocalCmd ( $sCmd,
                                   $sCmdArgs,
                                   \$iExitCode,
                                   $iTimeout,
                                   \@lsCmdOut ) ) {
      &logTrace ( 20083,
      "<%s> Command [%s] with args [%s] executed successfully with exit code [%s]",
      $sMyName, $sCmd, $sCmdArgs, $iExitCode );
    }
    elsif ( ! defined $iExitCode ) { # Timed out..
      $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Unknown..
      &logError ( 20084,
      "<%s> Command [%s] with args [%s] timed out. Returning [%s]",
      $sMyName, $sCmd, $sCmdArgs, $iReturnCode );
    }
    else { # Failed..
      &logError ( 20085,
      "<%s> Command [%s] with args [%s] execution failed with exit code [%s].
      Cmd output [%s]",
      $sMyName, $sCmd, $sCmdArgs, $iExitCode, join ( chr 012, @lsCmdOut ) );
      $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
    }
  }
  else { # Use socket(3socket) based approach..
    if ( Comms::CheckPort ( $sAdminHost,
                            $iAdminPort ) ) {
      &logTrace ( 20086,
      "<%s> Server listening on Host [%s] Port [%s]: Socket method",
      $sMyName, $sAdminHost, $iAdminPort );
    }
    else {
      &logError ( 20087,
      "<%s> Server NOT listening on Host [%s] Port [%s]: Socket method",
      $sMyName, $sAdminHost, $iAdminPort );
      $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
    }
  }

  return $iReturnCode;
}

=pod

 -------------------------------------------------------------------------
 Function : bool ExecAdmCmd ( IN String ServerRoot,
                              IN String SSLDbPasswd );

 Description : Starts/Stops a Sun ONE Administration Server

 Parameters :

 $sServerRoot     Absolute path to ServerRoot of Admin Server
 $sSSLDbPasswd    Encrypted password for trusted database using vcsencrypt

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub ExecAdmCmd {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sServerRoot,
       $sSSLDbPasswd ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20088, "Subroutine <%s> called:
  ServerRoot        [%s]
  SSLDbPasswd       [%s]",

  $sMyName, $sServerRoot, $sSSLDbPasswd );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = 1; # Success..

  #-------------------------
  # Define local variables..
  #-------------------------
  my %hAdminCmd = ( 'SunJESAdm::Start', 'start',
                    'SunJESAdm::Stop',  'stop',
                    'SunJESAdm::Kill',  'stop' ); 

  my ( $sCmd, $sCmdArgs, $sAction, $iExitCode, $bTimedOut,
       $iTillWhen, @lsCmdOut ) = ();

  my $sMyCaller = ( caller 1 )[3];   # Get calling function..
  my $sDecryptCmd = $sCmdVcsDecrypt; # Get global..

  #--------------------------------
  # Set appropriate admin command..
  #--------------------------------
  if ( $sAction = $hAdminCmd { $sMyCaller } ) {
    $sCmd = "$sServerRoot/${sAction}-admin"; 

    if ( ! -e $sCmd ) {
      &logError ( 20089, "<%s> Command [%s] to [%s] does not exist",
      $sMyName, $sCmd, $sAction );
      $iReturnCode = 0; # Failure..
    }
    elsif ( ! -x $sCmd ) {
      &logError ( 20090, "<%s> Command [%s] to [%s] not executable",
      $sMyName, $sCmd, $sAction );
      $iReturnCode = 0; # Failure..
    }
    else { 
      &logTrace ( 20091, "<%s> Using [%s] to [%s] the Administration Server",
      $sMyName, $sCmd, $sAction );
    }
  }
  else {
    &logError ( 20092, "<%s> Got unsupported call from sub-routine [%s]",
    $sMyName, $sMyCaller );
    $iReturnCode = 0; # Failure..
  }
 
  $iReturnCode || return $iReturnCode; # If failed, return..
 
  #------------------------------------
  # Protected database passwd defined..
  #------------------------------------
  if ( $sSSLDbPasswd =~ /\S/ && $sAction eq 'start' ) {
    if ( ! -e $sDecryptCmd ) {
      &logError ( 20093, "<%s> Decrypt utility [%s] does not exist",
      $sMyName, $sDecryptCmd );
      $iReturnCode = 0; # Failure..
    }
    elsif ( ! -x $sDecryptCmd ) {
      &logError ( 20094, "<%s> Decrypt utility [%s] not executable",
      $sMyName, $sDecryptCmd );
      $iReturnCode = 0; # Failure..
    }  

    $iReturnCode || return $iReturnCode; # If failed, return..
  
    &logTrace ( 20095, "<%s> Decrypt utility [%s] located",
    $sMyName, $sDecryptCmd );

    #----------------------------------------
    # Adjust cmd and cmd args appropriately..
    #----------------------------------------
    $sCmdArgs = "$sSSLDbPasswd | /bin/awk '{ print }' | $sCmd",
    $sCmd = "$sDecryptCmd";
  }

  $sCmdArgs = '' unless defined $sCmdArgs; # Prevent undefs..

  &logTrace ( 20096,
  "<%s> Executing command [%s] with args [%s] to [%s] the Administration Server",
  $sMyName, $sCmd, $sCmdArgs, $sAction );

  #--------------------
  # Get time bearings..
  #--------------------
  $iTillWhen = $^T + SunJESAdm::GetEngineTimeout ( 1 ) - time; # With margin..

  #----------------------
  # Execute the command..
  #----------------------
  if ( $iReturnCode = SunJESAdm::ExecLocalCmd ( $sCmd,
                                                $sCmdArgs,
                                                \$iExitCode,
                                                $iTillWhen,
                                                \@lsCmdOut ) ) {
    &logTrace ( 20097,
    "<%s> Command [%s] executed successfully with exit code [%s]",
    $sMyName, $sCmd, $iExitCode );
  }
  else { # Failed..
    &logError ( 20098,
    "<%s> Command [%s] execution failed with exit code [%s]. Cmd output [%s]",
    $sMyName, $sCmd, $iExitCode, join ( chr 012, @lsCmdOut ) );
    $iReturnCode = 0; # Failure..
  }
  
  return $iReturnCode;
}

=pod

 -------------------------------------------------------------------------
 Function : list GetPidFiles ( IN String ServerRoot );

 Description : Locates PID files for a Sun ONE Administration Server

 Parameters :

 $sServerRoot       Absolute path to ServerRoot of Admin Server

 Returns : List of PID files for the Administration Server
 -------------------------------------------------------------------------

=cut

sub GetPidFiles {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sServerRoot ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20099, "Subroutine <%s> called:
  ServerRoot        [%s]",
  $sMyName, $sServerRoot );

  #----------------------
  # Set local variables..
  #----------------------
  my ( $sTmp, $sPidFile, $sEntryPoint, @lsPidFiles ) = ();
  ( $sEntryPoint = $0 ) =~ s#.*/##;        # Basename..

  foreach $sTmp ( qw ( pid admin.pid ) ) {
    $sPidFile = "$sServerRoot/admin-serv/logs/$sTmp";

    #-----------------
    # Test each file..
    #-----------------
    if ( ! -e $sPidFile ) {
      if ( $sEntryPoint eq 'monitor' ) {
        &logTrace ( 20100, "<%s> Could not locate PID file [%s]",
        $sMyName, $sPidFile );
      }
      elsif ( $sEntryPoint ne 'online' ) {
        &logWarn ( 20101, "<%s> Could not locate PID file [%s]",
        $sMyName, $sPidFile );
      }
    }
    elsif ( ! -s $sPidFile ) {
      &logWarn ( 20102, "<%s> PID file [%s] is of zero bytes",
      $sMyName, $sPidFile );
    }
    else {
      &logTrace ( 20103,
      "<%s> Located PID file [%s] for Administration Server",
      $sMyName, $sPidFile );
      push ( @lsPidFiles, $sPidFile );
    }
  }

  return @lsPidFiles;
}

=pod
 -------------------------------------------------------------------------
 Function : int GetEngineTimeout ( IN Bool UseMargin );

 Description : Calculates engine timeout for entry point
 
 Parameters  :

 $bUseMargin      Flag to indicate if safety margin is requested

                  boolean value in the set of {0,1} where
                  0 No safety margin needed
                  1 Safety margin to be used
 
 Returns:  Time in secs.
 -------------------------------------------------------------------------
=cut

sub GetEngineTimeout {
  #----------------
  # Get Arguments..
  #----------------
  my ( $bUseMargin ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20104, "Subroutine <%s> called with:
  UseMargin  [%s]",
  $sMyName, $bUseMargin );

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sAttribute, $sCommand, $sCommandArgs,
       $iEngineTimeout, $iExitCode ) = ();

  my ( $sEntryPointName, $sResourceType ) = reverse $0 =~ m@([^/]+)@g;
  $sResourceType = 'SunJESAdm' if $sResourceType !~ /\w/; # Ensure..

  #----------------------------------
  # Grace margins from VCS timeouts..
  #----------------------------------
  my %hiGraceMargin = ( 'MonitorTimeout', 05,   # Seconds..
                        'OnlineTimeout',  10,
                        'OfflineTimeout', 10,
                        'CleanTimeout',   10 );

  #-----------------------------------------
  # Form query string based on entry point..
  #-----------------------------------------
  $sCommand = $sCmdHatype; # Get global..
  ( $sAttribute = $sEntryPointName ) =~ s/(.*)/\u$1Timeout/;
  $sCommandArgs = "-value $sResourceType $sAttribute";

  #----------------------------------
  # Fire command and collect output..
  #----------------------------------
  ( $iEngineTimeout = readpipe ( "$sCommand $sCommandArgs" ) ) =~ s/\n//;

  $iExitCode = $? >> 8;  # Get exit status..

  #--------------------
  # Check exit status..
  #--------------------
  if ( $iExitCode ) {
    &logTrace ( 20105,
    "<%s> Command [%s] with Args [%s] failed with exitcode [%s]",
    $sMyName, $sCmdHatype, $sCommandArgs, $iExitCode );
    $iEngineTimeout = $bUseMargin?$hiGraceMargin{$sAttribute}:0;
  }
  elsif ( ! $iEngineTimeout ) { # Empty value..
    &logTrace ( 20106, "<%s> Could not get [%s] from engine",
    $sMyName, $sAttribute );
    $iEngineTimeout = $bUseMargin?$hiGraceMargin{$sAttribute}:0;
  }
  else { # Success..
    &logTrace ( 20107, "<%s> Got [%s] as [%s] sec from engine",
    $sMyName, $sAttribute, $iEngineTimeout );
  }

  #----------------------------
  # Apply margin if requested.. 
  #----------------------------
  if ( $bUseMargin     &&                                 # Margin requested..
       $iEngineTimeout >= $hiGraceMargin{$sAttribute} ) { # And valid..
    $iEngineTimeout -= $hiGraceMargin{$sAttribute};       # Apply..

    &logTrace ( 20108,
    "<%s> Setting up [%s] sec margin. Returning [%s]",
    $sMyName, $hiGraceMargin{$sAttribute}, $iEngineTimeout );
  }

  return ( $iEngineTimeout ); # Secs..
}

=pod

 -------------------------------------------------------------------------
 Function : bool KillAdminProcs ( IN String ServerRoot );

 Description : Kills processes for a Sun ONE Administration Server

 Parameters :

 $sServerRoot       Absolute path to ServerRoot of Admin Server

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub KillAdminProcs {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sServerRoot ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20109, "Subroutine <%s> called:
  ServerRoot        [%s]",
  $sMyName, $sServerRoot );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1; # Success..

  #----------------------
  # Set local variables..
  #----------------------
  my ( @liGoners, @liUnDead ) = ();

  #--------------------------
  # If called during online..
  #--------------------------
  if ( ( caller 1 )[3] eq 'SunJESAdm::Start' ) {
    @liGoners = @lsRunningProcs; # Get global..
  }
  else {
    #----------------------------
    # Get processes for killing..
    #----------------------------
    SunJESAdm::GetProcState ( $sServerRoot, \@liGoners, 0 );
  }

  #------------------------------------------------------------------
  # We do a two level kill. The SIGINQ ensures  that  the  PID  is  a
  # valid process before we actually issue the SIGKILL. We don't want
  # to report a failure for SIGKILL in case  of  defunct  or  missing
  # processes.
  #------------------------------------------------------------------
  if ( scalar @liGoners ) {
    #----------------------
    # Failed to kill some..
    #----------------------
    if ( @liUnDead = grep {
                         kill ( 0, $_ ) && # Valid..
                       ! kill ( 9, $_ )    # and could NOT be killed..
                     } @liGoners ) {       # From what's found..
      &logError ( 20110,
      "<%s> PIDs not killed [%s] Total not killed [%s]",
      $sMyName, join ( ' ', @liUnDead ), scalar @liUnDead );
      $iReturnCode = 0; # Failure..
    }
    else { # All killed..
      &logInfo ( 20111,
      "<%s> Killed PIDs [%s] Total killed: [%s]",
      $sMyName, join ( ' ', @liGoners ), scalar @liGoners );
    }
  }
  else {
    &logTrace ( 20112,
    "<%s> Could not locate any process for ServerRoot [%s]",
    $sMyName, $sServerRoot );
  }

  return $iReturnCode;
}

=pod
 -------------------------------------------------------------------------
 Function : int ExternalMonitor ( IN String Cmd,
                                  IN String CmdArgs,
                                  IN Integer TillWhen );

 Description : Handles user defined external monitor program

 Parameters :

 $sCmd        Absolute path to external monitor program
 $sCmdArgs    Command line arguments, if any, for Cmd
 $iTillWhen   Represents the upper limit set by engine in EPOCH 

 Returns  : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------
=cut

sub ExternalMonitor {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sCmd,
       $sCmdArgs,
       $iTillWhen ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20113, "Subroutine <%s> called with:
  Cmd         [%s]
  CmdArgs     [%s]
  TillWhen    [%s]",

  $sMyName,
  $sCmd, $sCmdArgs, $iTillWhen );

  #------------------
  # Hopeful success..
  #------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE;

  #------------------------
  # Setup local variables..
  #------------------------
  my ( $iExitCode, $iTimeout, @lsCmdOut ) = ();

  #-------------------------
  # Handled exit code hash..
  #-------------------------
  my %hiExitCodes = (  0, VCSDef_MonitorExit::ONLINE,
                     110, VCSDef_MonitorExit::ONLINE,
                       1, VCSDef_MonitorExit::OFFLINE,
                     100, VCSDef_MonitorExit::OFFLINE );


  #---------------------------------------------
  # Check if command can be found and executed..
  #---------------------------------------------
  if ( ! -e $sCmd ) {
   &logError ( 20114,
    "<%s> External monitor [%s] does not exist",
    $sMyName, $sCmd );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( ! -x $sCmd ) {
   &logError ( 20115,
    "<%s> External monitor [%s] not executable.",
    $sMyName, $sCmd );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( ( $iTimeout = $iTillWhen - ( time + 5 ) ) <= 5 ) { # Lack of time..
   &logError ( 20116,
    "<%s> Insufficient time to proceed with external monitor [%s].
    Short by [%s] sec", $sMyName, $sCmd, abs ( $iTimeout ) );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE;  
  }

  #--------------------------------------------
  # Return if validation failed at this point..
  #--------------------------------------------
  unless ( VCSDef_MonitorExit::ONLINE == $iReturnCode ) {
    return $iReturnCode;
  }

  &logTrace ( 20117,
  "<%s> Executing external monitor with Command [%s] Args " .
  "[%s] timeout [%s] secs",
  $sMyName, $sCmd, $sCmdArgs, $iTimeout );

  #----------------------
  # Execute the command..
  #----------------------
  SunJESAdm::ExecLocalCmd ( $sCmd,
                            $sCmdArgs,
                            \$iExitCode,
                            $iTimeout,
                            \@lsCmdOut );

  #------------------------
  # Report on exit status..
  #------------------------
  if ( ! defined $iExitCode ) {
    &logError ( 20118,
    "<%s> Command [%s] with args [%s] timed out",
    $sMyName, $sCmd, $sCmdArgs );
    $iExitCode = 'undef';
  }
  elsif ( $hiExitCodes { $iExitCode } &&
          VCSDef_MonitorExit::ONLINE ==
          $hiExitCodes { $iExitCode } ) {
    &logTrace ( 20119,
    "<%s> Command [%s] with args [%s] executed successfully with exit code [%s].
    Cmd output [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iExitCode, join ( chr 012, @lsCmdOut ) );
  }
  else {
    &logWarn ( 20120,
    "<%s> Command [%s] with args [%s] execution failed with exit code [%s].
    Cmd output [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iExitCode, join ( chr 012, @lsCmdOut ) );
  }

  #--------------------------------
  # Setup appropriate return code..
  #--------------------------------
  $iReturnCode = $hiExitCodes { $iExitCode } ||
                 VCSDef_MonitorExit::UNKNOWN;

  return $iReturnCode;
}


=pod
 -------------------------------------------------------------------------
 Function : bool ExecLocalCmd ( IN String Command,
                                IN String CommandArgs,
                                OUT RefInt ExitCode,
                                IN Integer Timeout,
                                OUT ListReference CmdOut );

 Description : Wrapper function for
               Sys::RunWithEnvCmdWithOutputWithTimeOut to collect output
               of the command.
 
 Parameters :

 $sCommand         Absolute path to location of command
 $sCommandArgs     Arguments to command
 $riExitCode       Reference to collect exit code of command
 $iTimeout         Time limit to execute the command
 $rlsCmdOut        Stores o/p from command if successful

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------
=cut

sub ExecLocalCmd {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sCommand,
       $sCommandArgs,
       $riExitCode,
       $iTimeout,
       $rlsCmdOut ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20121, "Subroutine <%s> called with:
  Command     [%s]
  CommandArgs [%s]
  riExitCode  [%s]
  Timeout     [%s]
  rlsCmdOut   [%s]",

  $sMyName,
  $sCommand, $sCommandArgs, $riExitCode, $iTimeout, $rlsCmdOut );

  #---------------------------------------
  # Setup return code to hopeful success..
  #---------------------------------------
  my $iReturnCode = 1;

  #-----------------------------------
  # Verify timeout is greater than 1..
  #-----------------------------------
  if ( $iTimeout <= 1 ) {
    &logError ( 20122,
    "<%s> Insufficient time [%s] to execute Cmd [%s] with Args [%s].",
    $sMyName, $iTimeout, $sCommand, $sCommandArgs );
    return 0;
  }

  #-------------------
  # Set up arguments..
  #-------------------
  my $bSourceUserEnv = 0;   # Do not source..
  my $sEnvFile = "";        # No env file..
  my $sUser = 'root';       # Run as root..
  my $sTmpFile = tmpnam (); # Outfile..
  my $bTimedOut = 0;        # Timeout variable..

  @$rlsCmdOut = (); # Reset..

  $sCommandArgs .= " 1> $sTmpFile"; # Append..

  &logTrace ( 20123,
  "<%s> Executing Command [%s] with Args [%s]",
  $sMyName, $sCommand, $sCommandArgs );

  $iReturnCode = Sys::RunWithEnvCmdWithOutputWithTimeOut ( $sEnvFile,
                                                           $bSourceUserEnv,
                                                           $sCommand,
                                                           $sCommandArgs,
                                                           $riExitCode,
                                                           $sUser,
                                                           $rlsCmdOut,
                                                           \$bTimedOut,
                                                           $iTimeout );

  #--------------------
  # Check return code..
  #--------------------
  if ( $iReturnCode ) {   # No error with command execution..
    if ( $$riExitCode ) { # Failure..
      &logTrace ( 20124,
      "<%s> Command [%s] with args [%s] returned exit code [%s]",
      $sMyName, $sCommand, $sCommandArgs, $$riExitCode );
      $iReturnCode = 0; # Failure..
    }
    else {
      &logTrace ( 20125,
      "<%s> Command [%s] with args [%s] executed successfully with exit code [%s]",
      $sMyName, $sCommand, $sCommandArgs, $$riExitCode );
    }

    #-------------------------
    # Now collect the output..
    #-------------------------
    if ( -s $sTmpFile ) {
      if ( open ( LOGFILE, $sTmpFile ) ) {
        @$rlsCmdOut = grep { s/\n// || 1 } <LOGFILE>; # Read file..

        &logTrace ( 20126, "<%s> Command output :----------:\n[%s]",
        $sMyName, join ( chr 012, @$rlsCmdOut ) );

        close ( LOGFILE ) ||
        &logWarn ( 20127, "<%s> Could not close [%s] after read: [%s]",
        $sMyName, $sTmpFile, $! );
      }
      else {
        &logWarn ( 20128, "<%s> Could not open [%s] for logging: [$!]",
        $sMyName, $sTmpFile, $! );
      }
    }
    else {
      &logTrace ( 20129,
      "<%s> Could not collect output of command", $sMyName );
    }
  }
  else { # Execution error..
    if ( $bTimedOut ) {  # We timed out..
      &logError ( 20130,
      "<%s> Command [%s] with args [%s] did not complete within [%s] seconds",
      $sMyName, $sCommand, $sCommandArgs, $iTimeout );
    }
    else { # Failure due to something else..
      &logError ( 20131,
      "<%s> Failed in call to RunWithEnvCmdWithOutputWithTimeOut",
      $sMyName );
    }
    
    @$rlsCmdOut = (); # Reset..
  }

  #-----------
  # Clean up..
  #-----------
  unlink ( $sTmpFile ) ||
  &logWarn ( 20132, "<%s> Could not remove [%s] after read: [%s]",
  $sMyName, $sTmpFile, $! );

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool RunSecondLevelMonitor ( IN Integer SecondLevelMonitor,
                                         IN String ResourceType,
                                         IN String ResourceName );

 Description : Tells if its the correct cycle to perform second level
               monitoring.
 Parameters :

 $iSecondLevelMonitor   Cycle count to run in-depth monitoring
 $sResourceType         Type of VCS Resource
 $sResourceName         Name of VCS Resource

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub RunSecondLevelMonitor {
  #----------------
  # Get Arguments..
  #----------------
  my ( $iSecondLevelMonitor,
       $sResourceType,
       $sResourceName ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20133, "Subroutine <%s> called:
  SecondLevelMonitor        [%s]
  ResourceType              [%s]
  ResourceName              [%s]",

  $sMyName, $iSecondLevelMonitor, $sResourceType, $sResourceName );

  #---------------------------
  # Set Return Code to false..
  #---------------------------
  my $iReturnCode = 0; # Do not run SLM..

  #------------------------
  # Setup local variables..
  #------------------------
  my $sSLMDir = "/tmp/.VRTS$sResourceType";
  my $sSLMFile = "$sSLMDir/$sResourceName.slm";
  my $iSLMCntr = undef;

  #--------------------------------------------
  # Validate SecondLevelMonitor trivial cases..
  #--------------------------------------------
  if ( 0 == $iSecondLevelMonitor ||
       1 == $iSecondLevelMonitor ) {
    $iReturnCode = $iSecondLevelMonitor; # Assign..

    &logTrace ( 20134,
    "<%s> Got request to remove SLM file [%s]. Returning with [%s]",
    $sMyName, $sSLMFile, $iSecondLevelMonitor );

    unlink $sSLMFile if ( -e $sSLMFile ); # Remove SLM file..
    return $iReturnCode;
  }

  #----------------------------------
  # Create history folder if needed..
  #----------------------------------
  if ( -d $sSLMDir ) {
    &logTrace ( 20135, "<%s> SLM folder [%s] exists",
    $sMyName, $sSLMDir );
  }
  elsif ( ! mkdir ( $sSLMDir, 0700 ) ) {
    &logWarn ( 20136, "<%s> Failed to create SLM folder [%s]",
    $sMyName, $sSLMDir );
    return 1; # Run SLM..
  }
  
  #--------------------------
  # Create SLM history file..
  #--------------------------
  if ( sysopen ( SLMFILE, $sSLMFile, O_RDWR|O_SYNC|O_CREAT, 0600 ) ) {
    &logTrace ( 20137,
    "<%s> Opened SLM file [%s] successfully for read/write",
    $sMyName, $sSLMFile );
  }
  else {
    &logWarn ( 20138,
    "<%s> Failed to open SLM file [%s] for read/write",
    $sMyName, $sSLMFile );
    return 1; # Run SLM..
  }
  
  #---------------
  # Read in file..
  #---------------
  ( $iSLMCntr ) = grep { s/\n// } <SLMFILE>; 

  #-----------------
  # Corrupted file..
  #-----------------
  if ( ! defined $iSLMCntr || $iSLMCntr < 0 ) {

    $iSLMCntr ||= ''; # Avoid undefs..

    &logTrace ( 20139,
    "<%s> SLM file [%s] contains [%s]. Re-setting to [%s]",
    $sMyName, $sSLMFile, $iSLMCntr, $iSecondLevelMonitor );
    $iSLMCntr = $iSecondLevelMonitor; # Set to default..
  }
  
  #-----------------------------------------
  # Update counter, if less than requested..
  #-----------------------------------------
  if ( $iSecondLevelMonitor > $iSLMCntr ) {
    ++ $iSLMCntr;
  }
  else {
    $iSLMCntr = 1;    # Now's the time..
    $iReturnCode = 1; # Run SLM..
  }
  
  #------------------
  # Update the file..
  #------------------
  seek ( SLMFILE, 0, 0 ); # Rewind before update..

  &logTrace ( 20140,
  "<%s> Updating SLM file [%s] with [%s]. Return code is [%s]",
  $sMyName, $sSLMFile, $iSLMCntr, $iReturnCode );

  print SLMFILE "$iSLMCntr\n"; # Update..
  
  close ( SLMFILE ) ||
  &logWarn ( 20141,
  "<%s> Close failed on SLM file [%s] after update",
  $sMyName, $sSLMFile, $! );

  return $iReturnCode;
}

=pod

 -------------------------------------------------------------------------
 Function : bool IsIPLocal ( IN String Host );
                             

 Description : Tells if IP corresponding to virtual host is plumbed
               locally.
 Parameters :

 $sHost         Hostname/IPv4 address of virtual host

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub IsIPLocal {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sHost ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20142, "Subroutine <%s> called:
  Host            [%s]",

  $sMyName, $sHost );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #-----------------------------
  # Initialise local variables..
  #-----------------------------
  my ( $iAddr, $sName, $sStrAddr, $iExitCode, @lsCmd ) = ();

  my $sNetCmd = "/bin/netstat";
  my $sNetArgs = $^O eq 'linux'?'-eni':'-ian';

  #----------------------
  # Resolve his request..
  #----------------------
  if ( $iAddr = inet_aton $sHost ) {
    $sName = gethostbyaddr ( $iAddr, AF_INET );
    $sStrAddr = inet_ntoa $iAddr;

    &logTrace ( 20143,
    "<%s> Resolved [%s] to Host [%s] IPv4 Address [%s]",
    $sMyName, $sHost, $sName, $sStrAddr );
  }
  else {
    &logError ( 20144,
    "<%s> Could not resolve Host [%s]", $sMyName, $sHost );
    $iReturnCode = 0;
  }
  
  $iReturnCode || return $iReturnCode; # Return, if failed..

  #--------------------------
  # Get locally plumbed IPs..
  #--------------------------
  @lsCmd = grep { s/\n// || 1 }
           readpipe ( "$sNetCmd $sNetArgs" );
  
  #--------------------
  # Check exit status..
  #--------------------
  if ( $iExitCode = $? >> 8 ) {
    &logError ( 20145,
    "<%s> Cmd [%s] with Args [%s] failed with Exit Code [%s]",
     $sNetCmd, $sNetArgs, $iExitCode );
    $iReturnCode = 0;
  }
  elsif ( grep { /\b$sStrAddr\b/o } @lsCmd ) {
    &logTrace ( 20146,
    "<%s> IP Address [%s] is plumbed locally",
    $sMyName, $sStrAddr );
  }
  else {
    &logTrace ( 20147,
    "<%s> IP Address [%s] is not plumbed locally",
    $sMyName, $sStrAddr );
    $iReturnCode = 0;
  }

  return $iReturnCode;
}

1; # So PERL knows its included..
